/*
 * navu-gpioirq.c - GPIO driver for "nav-u" NAVU/4 series
 *
 *     GPIO-IRQ handler and character device interface.
 *
 * Copyright (c) 2007 Sony Corporation.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/list.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/mach/irq.h>
#include <asm/arch/system.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>
#include <asm/arch/xyz3-devnum.h>

#include <asm/arch/navu-gpio.h>



/*
 * Device file
 */
static int DevMajorNumber = XYZ3_MAJOR_GPIO;
static int DevMinorNumber = XYZ3_MINOR_GPIO;
static int DevCount = 1;
static const char *DevName = "gpio";


static irqreturn_t gpio_int( int irq_nr, void *dev_id, struct pt_regs *regs );


/****************************************************************************
 *                      IRQ Management Routines
 ****************************************************************************/

LIST_HEAD(IrqList);
typedef struct {
    struct list_head  list;
    unsigned long     port;
    int               irq_nr;
    wait_queue_head_t *wait_q;
    atomic_t          fall_count;   /* These counters used to ensure sleeping process... */
    atomic_t          rise_count;   /* ...to be awaken AFTER the interrupt received. */
    atomic_t          cancel_count; /* Used for canceling waiting threads */
    atomic_t          refcount;     /* if zero, you can remove this entry */
} navu_irq_entry;

DECLARE_MUTEX( IRQListSem );
static spinlock_t IRQListLock = SPIN_LOCK_UNLOCKED;
static unsigned long IRQListLockFlags;

/**
 * irq_entry_free - free the memory region of navu_irq_entry*
 * @entry pointer to the memory region you want to free
 */
static void
irq_entry_free(navu_irq_entry* entry)
{
    if( entry != NULL ) {
        if( entry->wait_q != NULL ) {
            kfree( entry->wait_q );
            entry->wait_q = NULL;
        }
        kfree( entry );
        entry = NULL;
    }
    return;
}


/**
 * irq_entry_alloc - allocate a memory region for navu_irq_entry*
 */
static navu_irq_entry*
irq_entry_alloc(void)
{
    navu_irq_entry* entry;

    entry = kmalloc(sizeof(navu_irq_entry), GFP_KERNEL);
    if( entry == NULL ) {
        PRINTK_ERR("kmalloc() FAILED for navu_irq_entry");
        goto InitErr;
    }

    entry->wait_q = kmalloc(sizeof(wait_queue_head_t), GFP_KERNEL);
    if( entry->wait_q == NULL ) {
        PRINTK_ERR("kmalloc() FAILED for wait_queue_head_t ");
        goto InitErr;
    }

    return entry;

InitErr:
    irq_entry_free(entry);
    return NULL;
}


/**
 * irq_entry_get - Pick an entry for GPIO 'port' and register.
 * @port GPIO port name to register.
 */
static navu_irq_entry*
irq_entry_get(unsigned long port)
{
    int rc, irq_nr;
    navu_irq_entry* entry;

    irq_nr = port2irqnum(port);

    DPRINTK("Request for navu_irq_entry for port=0x%lX, irq_nr=%d", port, irq_nr);

    if ( down_interruptible(&IRQListSem) ) {
        return NULL;
    }
    list_for_each_entry(entry, &IrqList, list) {
        if ( entry->irq_nr == irq_nr ) {
            DPRINTK("Found already registered entry");
            atomic_inc(&entry->refcount);
            up(&IRQListSem);
            return entry;
        }
    }

    /*
     * In case the port is new to us.
     */
    DPRINTK("No previously registered entry. calling irq_entry_alloc().");
    entry = irq_entry_alloc();
    if ( ! entry ) {
        up( &IRQListSem );
        return NULL;
    }

    init_waitqueue_head( entry->wait_q );
    entry->port   = port;
    entry->irq_nr = irq_nr;
    set_irq_type( irq_nr, (IRQT_FALLING|IRQT_RISING) );
    rc = request_irq( irq_nr, gpio_int, SA_INTERRUPT, DevName, NULL);
    if ( rc < 0 ) {
        PRINTK_ERR("request_irq() FAILED (%d)", rc);
        up( &IRQListSem );
        irq_entry_free(entry);
        return NULL;
    }

    atomic_set(&entry->refcount, 1);
    atomic_set(&entry->fall_count, 0);
    atomic_set(&entry->rise_count, 0);
    atomic_set(&entry->cancel_count, 0);
    spin_lock_irqsave(&IRQListLock, IRQListLockFlags);
    list_add(&entry->list, &IrqList);
    spin_unlock_irqrestore(&IRQListLock, IRQListLockFlags);

    up( &IRQListSem );
    return entry;
}


/**
 * irq_entry_release - Release a registered irq entry.
 * @port GPIO port name to release.
 */
static void
irq_entry_release(unsigned long port)
{
    int irq_nr;
    navu_irq_entry* entry;
    navu_irq_entry* tmp;

    irq_nr = port2irqnum(port);
    DPRINTK("Release request for navu_irq_entry: port=0x%lX, irq_nr=%d", port, irq_nr);

    down(&IRQListSem);

    list_for_each_entry_safe(entry, tmp, &IrqList, list) {
        if ( entry->irq_nr == irq_nr ) {
            if ( atomic_dec_and_test(&entry->refcount) ) {
                DPRINTK("The irq entry no more refered (I'm the last one.)");
                free_irq( irq_nr, NULL );
                spin_lock_irqsave(&IRQListLock, IRQListLockFlags);
                list_del( &entry->list );
                spin_unlock_irqrestore(&IRQListLock, IRQListLockFlags);
                irq_entry_free( entry );
            }
            else {
                DPRINTK("The irq entry is still in use. Not released.");
            }
            up(&IRQListSem);
            return;
        }
    }

    PRINTK_ERR("Entry for IRQ#%d NOT FOUND "
               "get/release mismatch?", irq_nr);
    up(&IRQListSem);
    return;
}


/**
 * gpio_int - Interrupt handler for GPIO Interrupt
 */
static irqreturn_t
gpio_int( int irq_nr, void *dev_id, struct pt_regs *regs )
{
    navu_irq_entry* entry;

    spin_lock_irqsave(&IRQListLock, IRQListLockFlags);
    list_for_each_entry(entry, &IrqList, list) {
        if ( entry->irq_nr == irq_nr ) {
            DPRINTK("Interrupt received for IRQ#%d, level=%d",
                    irq_nr, navuGpioGetLevel(entry->port) );
            if ( navuGpioGetLevel(entry->port) ) {
                atomic_inc(&entry->rise_count);
            }
            else {
                atomic_inc(&entry->fall_count);
            }
            wake_up_interruptible( entry->wait_q );
            spin_lock_irqsave(&IRQListLock, IRQListLockFlags);
            return IRQ_HANDLED;
        }
    }

    PRINTK_ERR("Unknown IRQ#%d received (ignored).", irq_nr);
    spin_lock_irqsave(&IRQListLock, IRQListLockFlags);
    return IRQ_HANDLED;
}


/**
 * navuGpioGetWait - IOCTL Backend function for IRQ request.
 * @arg port and trigger for wait.
 */
int
navuGpioIrqWait(navu_irq_ioctl_t* arg)
{
    int rc, irq_nr, portlevel;
    navu_irq_entry* entry;
    unsigned long port;
    int fall_count_before, rise_count_before, cancel_count_before;
    int trigger;

    port    = arg->port;
    trigger = arg->trigger;
    irq_nr  = port2irqnum(port);
    portlevel = navuGpioGetLevel( port );

    DPRINTK("Waiting for port=0x%lX, irq_nr=%ld, trigger=0x%X",
            port, port2irqnum(port), trigger);

    if( ( (trigger == NAVU_IRQ_LOW_LEVEL)  && (portlevel == 0) )
     || ( (trigger == NAVU_IRQ_HIGH_LEVEL) && (portlevel == 1) ) )
    {
        DPRINTK("Level trigger; returning befor wait_event_interruptible().");
        return portlevel;
    }

    entry = irq_entry_get( port );
    if( ! entry ) {
        return -ENOMEM;
    }

    /*
     * The interrupt handler increments these values accourding to the
     * refering GPIO status just after an interrupt.
     * If we found these values
     */
    fall_count_before   = atomic_read(&entry->fall_count);
    rise_count_before   = atomic_read(&entry->rise_count);
    cancel_count_before = atomic_read(&entry->cancel_count);

    switch (trigger) {
        case NAVU_IRQ_LOW_LEVEL:
        case NAVU_IRQ_FALL_EDGE:
            rc = wait_event_interruptible(
                    *(entry->wait_q),
                    (  fall_count_before < atomic_read(&entry->fall_count)
                    || cancel_count_before < atomic_read(&entry->cancel_count) )
                 );
            break;

        case NAVU_IRQ_HIGH_LEVEL:
        case NAVU_IRQ_RISE_EDGE:
            rc = wait_event_interruptible(
                    *(entry->wait_q),
                    (  rise_count_before < atomic_read(&entry->rise_count)
                    || cancel_count_before < atomic_read(&entry->cancel_count) )
                 );
            break;

        case NAVU_IRQ_DOUBLE_EDGE:
            rc = wait_event_interruptible(
                    *(entry->wait_q),
                    (  fall_count_before < atomic_read(&entry->fall_count)
                    || rise_count_before < atomic_read(&entry->rise_count)
                    || cancel_count_before < atomic_read(&entry->cancel_count) )
                 );
            break;

        default:
            PRINTK_ERR("Invalid trigger type %d", trigger);
            irq_entry_release( port );
            return -EINVAL;
    }

    irq_entry_release( port );

    if( rc != 0 ) {
        DPRINTK("IRQ wake awaken by signal.");
        return -ERESTARTSYS;
    }
    else
    if( cancel_count_before < atomic_read(&entry->cancel_count) ) {
        /* There might be a chance that &entry->cancel_count have been
         * incremented just after comparison in wait_event_interruptible()
         * and before the if() above. But we leave this window since
         * it won't break semantics.
         */
        DPRINTK("IRQ wait canceled.");
        return -ECANCELED;
    } else {
        return navuGpioGetLevel( port );
    }
}


int
navuGpioIrqCancel(unsigned long port)
{
    navu_irq_entry* entry;
    int irq_nr = port2irqnum(port);


    if ( down_interruptible(&IRQListSem) ) {
        return -EINTR;
    }
    list_for_each_entry(entry, &IrqList, list) {
        if ( entry->irq_nr == irq_nr ) {
            atomic_inc(&entry->cancel_count);
            wake_up_interruptible( entry->wait_q );
            up(&IRQListSem);
        }
    }
    DPRINTK("Attempted thread cancel but waiting thread for IRQ#%d.", irq_nr);
    up(&IRQListSem);
    return 0;
}

/****************************************************************************
                     Kernel File Operations Interface
 ****************************************************************************/

static int gpio_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    int stat = 0;

    switch( cmd )
    {
        case NAVUGP_GET_LEVEL:
            stat = navuGpioGetLevel(arg);
            break;
        case NAVUGP_SET_HIGH:
            stat = navuGpioSetLevel(arg, 1);
            break;
        case NAVUGP_SET_LOW:
            stat = navuGpioSetLevel(arg, 0);
            break;
        case NAVUGP_IRQ_WAIT:
            stat = navuGpioIrqWait( (navu_irq_ioctl_t*)arg );
            break;
        case NAVUGP_IRQ_CANCEL:
            stat = navuGpioIrqCancel(arg);
            break;
        default:
            stat = -EINVAL;
    }
    return stat;
}


static struct file_operations gpio_fops = {
    owner:   THIS_MODULE,
    ioctl:   gpio_ioctl,
};


/****************************************************************************
                    Kernel Power Management Interface
 ****************************************************************************/


/* Save all GPIO state before suspend and restore at resume */
static struct {
    u32     PinLevel0;
    u32     PinLevel1;
    u32     PinLevel2;
    u32     PinLevel3;
} SaveReg;


/*
 *  Power manager Support Routines
 */
static int __init gpio_probe(struct platform_device *dev)
{
    memset( &SaveReg, 0, sizeof( SaveReg ) );
    return 0;
}

static int gpio_remove(struct platform_device *dev)
{
    return 0;
}

static int gpio_suspend(struct platform_device *dev, pm_message_t state)
{
    SaveReg.PinLevel0 = GPLR0;
    SaveReg.PinLevel1 = GPLR1;
    SaveReg.PinLevel2 = GPLR2;
    SaveReg.PinLevel3 = GPLR3;

    return 0;
}

static int gpio_resume(struct platform_device *dev)
{
    GPSR0 =  SaveReg.PinLevel0;
    GPCR0 = ~SaveReg.PinLevel0;
    GPSR1 =  SaveReg.PinLevel1;
    GPCR1 = ~SaveReg.PinLevel1;
    GPSR2 =  SaveReg.PinLevel2;
    GPCR2 = ~SaveReg.PinLevel2;
    GPSR3 =  SaveReg.PinLevel3;
    GPCR3 = ~SaveReg.PinLevel3;

    return 0;
}


static struct platform_driver navugpio_driver = {
    .probe      = gpio_probe,
    .remove     = gpio_remove,
    .suspend    = gpio_suspend,
    .resume     = gpio_resume,
    .driver     = {
        .name   = "navu-gpio",
    },
};




/****************************************************************************
                    Kernel Init / Cleanup Routines
 ****************************************************************************/

static struct cdev gpio_cdev;
static int __init gpio_init( void )
{
    int result;
    int err;
    int devno;
    dev_t dev = 0;

    DPRINTK("nav-u GPIO driver started (%s %s) for "
#ifdef MACH_XYZ3
    "XYZ3"
#endif
#ifdef MACH_XYZ4
    "XYZ4"
#endif
    , __DATE__, __TIME__);

    /* Actually allocate io-port region. Always success. */
    if( request_mem_region(GPIO_BASE_ADDR, GPIO_SIZE, DevName) == NULL )
    {
        PRINTK_ERR("request_mem_region() error!!");
        return -EBUSY;
    }

    /* Append this device driver to system. */
    if (DevMajorNumber) {
        dev = MKDEV(DevMajorNumber, DevMinorNumber);
        result = register_chrdev_region(dev, DevCount, DevName);
    }
    else {
        result = alloc_chrdev_region(&dev, DevMinorNumber, DevCount, DevName);
        DevMajorNumber = MAJOR(dev);
    }
    if (result < 0) {
        PRINTK_ERR("can't get major %d", DevMinorNumber);
        return result;
    }

    devno = MKDEV(DevMajorNumber, DevMinorNumber);
    cdev_init(&gpio_cdev, &gpio_fops);
    gpio_cdev.owner = THIS_MODULE;
    gpio_cdev.ops = &gpio_fops;

    err = cdev_add(&gpio_cdev, devno, 1);
    if( err ) {
        PRINTK_ERR("Error %d while cdev_add()", err);
    }

    result = platform_driver_register(&navugpio_driver);
    if( result != 0 ) {
        PRINTK_ERR("driver_register() failed. (%d)", result);
    }

    return result;
}


static void __exit gpio_cleanup(void)
{
    dev_t devno = MKDEV(DevMajorNumber, DevMinorNumber);

    if ( IrqList.prev != IrqList.next ) {
        PRINTK_ERR("Oops, IrqList is not empty!");
    }

    platform_driver_unregister(&navugpio_driver);

    cdev_del(&gpio_cdev);

    release_mem_region(GPIO_BASE_ADDR, GPIO_SIZE);
    unregister_chrdev_region(devno, DevCount);
}


module_init(gpio_init);
module_exit(gpio_cleanup);

EXPORT_SYMBOL(navuGpioIrqWait);
EXPORT_SYMBOL(navuGpioIrqCancel);

MODULE_LICENSE("GPL");


