/*
 * xyz3-accsync.c - A very tiny driver for ACC synchronized power on / off.
 *
 *                           Author: Ryoichi KATO <Ryoichi.Kato@jp.sony.com>
 *                           Last Change: 2007/01/09 18:48:10.
 *                           Copyright: Sony Corporation.
 *
 * mknod /dev/accsync c 250 0
 *    Enable / disable ACC (car ACCessory power supply) synchronized
 *    power on feature (PXA27x's PWER/PFER register) which is connected to.
 *    PXA's GPIO-1
 *
 *      - Write a "0" char or a byte 0 into device file to disable ACC-sync.
 *      - Write a "1" char or a byte 1 into device file to enable ACC-sync.
 *      - Read the device file to see current status of the ACC-sync configuration.
 *        (Returns "0\n" if disabled, "1\n" if enabled)
 *
 *
 * mknod /dev/wakeuptrig c 250 1
 *  - Read the device to see wakeup trigger bit of the previous boot / resume.
 *    The bits are cleared at suspend and software-reboot (both uses DeepSleep).
 *    Returns raw bits of PEDR register. The valid bits for "nav-u" XYZ-3 are:
 *
 *      0x80000000  EDRTC  RTC-Alarm wakeup (Used by software reset)
 *      0x00000002  ED1    GPIO-1; Cradle ACC-ON
 *      0x00000001  ED0    GPIO-0; POWER button
 *
 *
 * NOTE: Device's major / minor numbers are defined in <asm/arch/xyz3-devnum.h>
 *
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <asm/uaccess.h> /* copy_to_user(), copy_from_user() */

#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>
#include <asm/arch/xyz3-devnum.h>

/****************************************************************************/
/*
 * Prototypes
 */
static int pw_open(struct inode * inode, struct file * filp);

static ssize_t accsync_read(struct file * file, char __user * buf,
                            size_t count, loff_t *ppos);
static ssize_t accsync_write(struct file * file, const char __user * buf,
                             size_t count, loff_t *ppos);
static ssize_t wakeuptrig_read(struct file * file, char __user * buf,
                            size_t count, loff_t *ppos);
int accsync_getstat(void);
void accsync_enable(void);
void accsync_disable(void);


/****************************************************************************/
/*
 * Global data
 */
static struct file_operations pw_fops = {
    .open       = pw_open,
};

static struct file_operations accsync_fops = {
    .read       = accsync_read,
    .write      = accsync_write,
};

static struct file_operations wakeuptrig_fops = {
    .read       = wakeuptrig_read,
};


/****************************************************************************/
/*
 * Functions
 */
static int pw_open(struct inode * inode, struct file * filp)
{
    switch (iminor(inode)) {
        case XYZ3_MINOR_ACCSYNC:
            filp->f_op = &accsync_fops;
            break;

        case XYZ3_MINOR_ACCSYNC_WAKEUPTRIGER:
            filp->f_op = &wakeuptrig_fops;
            break;

        default:
            return -ENXIO;
    }
    return 0;
}


static ssize_t accsync_read(struct file * file, char __user * buf,
                            size_t count, loff_t *ppos)
{
    if ( *ppos != 0 ) {
        /* already written; return EOF */
        return 0;
    }
    else {
        if ( count >= 2 ) {
            if ( accsync_getstat() ) {
                copy_to_user(buf, "1\n", 2);
            }
            else {
                copy_to_user(buf, "0\n", 2);
            }
            *ppos = 2;
            return 2;
        }
        else {
            return 0;
        }
    }
}

static ssize_t accsync_write(struct file * file, const char __user * buf,
                             size_t count, loff_t *ppos)
{
    char read_data;

    if ( count < 1 ) {
        return -EINVAL;
    }

    /*
     * Only look at the very beggining one byte of the data.
     */
    copy_from_user(&read_data, buf, 1);

    if ( read_data == 0 || read_data == '0' ) {
        accsync_disable();
        printk(KERN_INFO "accsync: disabled\n");
        return count;
    }
    else
    if ( read_data == 1 || read_data == '1' ) {
        accsync_enable();
        printk(KERN_INFO "accsync: enabled\n");
        return count;
    }
    else {
        return -EINVAL;
    }
}

static ssize_t wakeuptrig_read(struct file * file, char __user * buf,
                               size_t count, loff_t *ppos)
{
    if ( *ppos != 0 ) {
        /* already written; return EOF */
        return 0;
    }
    else {
        /*
        char result[2];
        result[0] = '0' + (PEDR & 0x3);
        result[1] = '\n';
        copy_to_user(buf, result, 2);
        *ppos = 2;
        */
        unsigned long pedr_value = PEDR;
        copy_to_user(buf, &pedr_value, sizeof(PEDR));
        *ppos = sizeof(PEDR); /* set file opsition accordingly */
        return sizeof(PEDR);
    }
}

/****************************************************************************/

int accsync_getstat(void)
{
    return ( (PWER & PFER & 0x02) >> 1 );
}

EXPORT_SYMBOL(accsync_getstat);

void accsync_enable(void)
{
    PWER  |=  0x02;  // Enable wake-up due to GPIO-1 edge detect.
    PRER  &= ~0x02;  // Disable GPIO-1 rising edge detect.
    PFER  |=  0x02;  // Enable  GPIO-1 falling edge detect.
}

void accsync_disable(void)
{
    PWER  &= ~0x02;  // Disable wake-up due to GPIO-1 edge detect.
}


/****************************************************************************/

/*
 * Initialize / Finalize Routines
 */
static void __exit accsync_exit( void )
{
    unregister_chrdev(XYZ3_MAJOR_ACCSYNC,"accsync");
}

static int __init accsync_init( void )
{
    printk("accsync: started. current status = %d\n", accsync_getstat() );
    if (register_chrdev(XYZ3_MAJOR_ACCSYNC,"accsync",&pw_fops)) {
        printk(KERN_ERR "accsync: Unable to get major %d for memory devs.\n",
               XYZ3_MAJOR_ACCSYNC);
        return -ENODEV;
    }
    return 0;
}


module_init(accsync_init);
module_exit(accsync_exit);

