/*
 * linux/arch/arm/mach-pxa/xyz3-ostimer.c
 *     -- Operating System Timers driver on PXA270
 *        
 * Author:     Hirokazu ABE and Ryoichi KATO
 * Created:    Feb 16, 2006
 * Copyright:  Sony Corp.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
/* #include <linux/malloc.h> */
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>

#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/mach/irq.h>

#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>

#include <asm/arch/xyz3-ostimer.h>
#include <asm/arch/xyz3-devnum.h>

#define OSTIMER_CH_MIN 4
#define OSTIMER_CH_MAX 11
#define OSTIMER_VALID_BITMASK (((2 << OSTIMER_CH_MAX) - 1) & ~((1 << OSTIMER_CH_MIN) - 1))

struct {
    unsigned long oscr_addr;
    unsigned long osmr_addr;
    unsigned long omcr_addr;
} static const OSTIMER_REGS[OSTIMER_CH_MAX+1] = {
    { 0x40A00010, 0x40A00000, 0, }, /* OS timer 0 can't use */
    { 0x40A00010, 0x40A00004, 0, }, /* OS timer 1 can't use */
    { 0x40A00010, 0x40A00008, 0, }, /* OS timer 2 can't use */
    { 0x40A00010, 0x40A0000C, 0, }, /* OS timer 3 can't use */
    { OSCR4_ADDR, OSMR4_ADDR, OMCR4_ADDR, },
    { OSCR5_ADDR, OSMR5_ADDR, OMCR5_ADDR, },
    { OSCR6_ADDR, OSMR6_ADDR, OMCR6_ADDR, },
    { OSCR7_ADDR, OSMR7_ADDR, OMCR7_ADDR, },
    { OSCR8_ADDR, OSMR8_ADDR, OMCR8_ADDR, },
    { OSCR9_ADDR, OSMR9_ADDR, OMCR9_ADDR, },
    { OSCR10_ADDR, OSMR10_ADDR, OMCR10_ADDR, },
    { OSCR11_ADDR, OSMR11_ADDR, OMCR11_ADDR, },
};

void xyz3_ostimer_setting(int ch, unsigned long count, int oneshot, int reset_on_match, int resolution)
{
    /* OMCR -> OSMR -> OSCR writing order is important. */
    __REG(OSTIMER_REGS[ch].omcr_addr) = (1 << 7) | (((~oneshot) & 1) << 6) | ((reset_on_match & 1) << 3) | (resolution & 0x7);
    __REG(OSTIMER_REGS[ch].osmr_addr) = count;
    __REG(OSTIMER_REGS[ch].oscr_addr) = 0;
}
EXPORT_SYMBOL(xyz3_ostimer_setting);

unsigned long xyz3_ostimer_get_count(int ch)
{
    return __REG(OSTIMER_REGS[ch].oscr_addr);
}
EXPORT_SYMBOL(xyz3_ostimer_get_count);

int xyz3_ostimer_get_resolution(int ch)
{
    return __REG(OSTIMER_REGS[ch].omcr_addr) & 0x7;
}
EXPORT_SYMBOL(xyz3_ostimer_get_resolution);

static DECLARE_WAIT_QUEUE_HEAD(ostimer_blockcall_wakeup_wq);

static unsigned long ostimer_blockcall_wakeup = 0;
static unsigned long ostimer_blockcall_mask = 0;
static spinlock_t ostimer_irq_lock = SPIN_LOCK_UNLOCKED;

int xyz3_ostimer_blockcall(int ch, unsigned long count, int resolution)
{
    unsigned long irqsave;
    int ret = 0;

    if((ch < OSTIMER_CH_MIN) || (OSTIMER_CH_MAX < ch)) return 0;

    if(test_and_set_bit(ch, &ostimer_blockcall_mask) == 0){

        /* start count and enable interrput */
        xyz3_ostimer_setting(ch, count, 1, 0, resolution);

        spin_lock_irqsave(&ostimer_irq_lock, irqsave);
        OIER |= (1 << ch);
        spin_unlock_irqrestore(&ostimer_irq_lock, irqsave);

        wait_event_interruptible(ostimer_blockcall_wakeup_wq, test_and_clear_bit(ch, &ostimer_blockcall_wakeup) != 0);
        clear_bit(ch, &ostimer_blockcall_mask);

        /* disable interrput and stop count */
        spin_lock_irqsave(&ostimer_irq_lock, irqsave);
        OIER &= ~(1 << ch);
        spin_unlock_irqrestore(&ostimer_irq_lock, irqsave);

        xyz3_ostimer_setting(ch, 0, 1, 0, OSTIMER_RESOLUTION_STOP);

        ret = 1;
    }

    return ret;
}
EXPORT_SYMBOL(xyz3_ostimer_blockcall);

static void (*ostimer_callback_tbl[OSTIMER_CH_MAX+1])(void) = {
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
};
static DECLARE_MUTEX(ostimer_callback_mutex);

static void ostimer_callback_handler(unsigned long unused);
static DECLARE_TASKLET(ostimer_callback_task, ostimer_callback_handler, 0);

static unsigned long ostimer_callback_wakeup = 0;
static unsigned long ostimer_callback_mask = 0;

int xyz3_ostimer_register_callback(int ch, unsigned long count, int resolution, void (*callback)(void))
{
    unsigned long irqsave;
    int ret = 0;

    if((ch < OSTIMER_CH_MIN) || (OSTIMER_CH_MAX < ch)) return 0;
    if(0 != down_interruptible(&ostimer_callback_mutex) ) return 0;

    tasklet_disable(&ostimer_callback_task);
    if(ostimer_callback_tbl[ch] == NULL){
        ostimer_callback_tbl[ch] = callback;
        set_bit(ch, &ostimer_callback_mask);

        /* start count and enable interrput */
        /* set match count - 1, included zero */
        xyz3_ostimer_setting(ch, count-1, 0, 1, resolution);

        spin_lock_irqsave(&ostimer_irq_lock, irqsave);
        OIER |= (1 << ch);
        spin_unlock_irqrestore(&ostimer_irq_lock, irqsave);

        ret = 1;
    }
    tasklet_enable(&ostimer_callback_task);

    up(&ostimer_callback_mutex);

    return ret;
}
EXPORT_SYMBOL(xyz3_ostimer_register_callback);

int xyz3_ostimer_remove_callback(int ch, void (*callback)(void))
{
    unsigned long irqsave;
    int ret = 0;

    if((ch < OSTIMER_CH_MIN) || (OSTIMER_CH_MAX < ch)) return 0;
    if(0 != down_interruptible(&ostimer_callback_mutex) ) return 0;

    tasklet_disable(&ostimer_callback_task);
    if(ostimer_callback_tbl[ch] == callback){
        clear_bit(ch, &ostimer_callback_mask);

        /* disable interrput and stop count */
        spin_lock_irqsave(&ostimer_irq_lock, irqsave);
        OIER &= ~(1 << ch);
        spin_unlock_irqrestore(&ostimer_irq_lock, irqsave);

        xyz3_ostimer_setting(ch, 0, 0, 1, OSTIMER_RESOLUTION_STOP);

        ostimer_callback_tbl[ch] = NULL;
        ret = 1;
    }
    tasklet_enable(&ostimer_callback_task);

    up(&ostimer_callback_mutex);

    return ret;
}
EXPORT_SYMBOL(xyz3_ostimer_remove_callback);

static void ostimer_callback_handler(unsigned long unused)
{
    int i;

    for(i = OSTIMER_CH_MIN; i <= OSTIMER_CH_MAX; i++){
        if(test_and_clear_bit(i, &ostimer_callback_wakeup) != 0){
            if(ostimer_callback_tbl[i] != NULL){
                (*(ostimer_callback_tbl[i]))();
            }
        }
    }
}

/* interval timer for user application */
static unsigned long ostimer_interval_wakeup = (unsigned long)-1; /* 0xFF.... */
static DECLARE_WAIT_QUEUE_HEAD(ostimer_interval_wakeup_wq);

static void ostimer_interval_handler(int ch)
{
    set_bit(ch, &ostimer_interval_wakeup);
    wake_up_interruptible(&ostimer_interval_wakeup_wq);
}

static void ostimer_interval_handler_ch4(void) { ostimer_interval_handler(4); }
static void ostimer_interval_handler_ch5(void) { ostimer_interval_handler(5); }
static void ostimer_interval_handler_ch6(void) { ostimer_interval_handler(6); }
static void ostimer_interval_handler_ch7(void) { ostimer_interval_handler(7); }
static void ostimer_interval_handler_ch8(void) { ostimer_interval_handler(8); }
static void ostimer_interval_handler_ch9(void) { ostimer_interval_handler(9); }
static void ostimer_interval_handler_ch10(void) { ostimer_interval_handler(10); }
static void ostimer_interval_handler_ch11(void) { ostimer_interval_handler(11); }

static void (* const ostimer_interval_handler_tbl[OSTIMER_CH_MAX+1])(void) = {
    NULL, NULL, NULL, NULL, /* CH0 - 3 is unused */
    &ostimer_interval_handler_ch4,
    &ostimer_interval_handler_ch5,
    &ostimer_interval_handler_ch6,
    &ostimer_interval_handler_ch7,
    &ostimer_interval_handler_ch8,
    &ostimer_interval_handler_ch9,
    &ostimer_interval_handler_ch10,
    &ostimer_interval_handler_ch11,
};

/* interrupt */

static irqreturn_t ostimer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    unsigned long irqsave;
    unsigned long ossr_req;
    irqreturn_t ret;

    spin_lock_irqsave(&ostimer_irq_lock, irqsave);

    ossr_req = OSSR & OSTIMER_VALID_BITMASK;
    if(ossr_req == 0){
        ret = IRQ_NONE;
    }
    else{
        OSSR = ossr_req; /* clear OSSR */

        ostimer_blockcall_wakeup = (ostimer_blockcall_wakeup | ossr_req) & ostimer_blockcall_mask;
        if(ostimer_blockcall_wakeup != 0){
            wake_up_interruptible(&ostimer_blockcall_wakeup_wq);
        }
        ostimer_callback_wakeup = (ostimer_callback_wakeup | ossr_req) & ostimer_callback_mask;
        if(ostimer_callback_wakeup != 0){
            tasklet_schedule(&ostimer_callback_task);
        }

        ret = IRQ_HANDLED;
    }

    spin_unlock_irqrestore(&ostimer_irq_lock, irqsave);

    return ret;
}

#ifdef CONFIG_PM

static unsigned long keep_oscr[OSTIMER_CH_MAX+1];
static unsigned long keep_osmr[OSTIMER_CH_MAX+1];
static unsigned long keep_omcr[OSTIMER_CH_MAX+1];
static unsigned long keep_oier;

static int ostimer_suspend(struct platform_device *dev, pm_message_t state)
{
    unsigned long irqsave;
    int i;

    spin_lock_irqsave(&ostimer_irq_lock, irqsave);
    keep_oier = OIER & OSTIMER_VALID_BITMASK;
    OIER &= ~OSTIMER_VALID_BITMASK;
    spin_unlock_irqrestore(&ostimer_irq_lock, irqsave);

    for(i = OSTIMER_CH_MIN; i <= OSTIMER_CH_MAX; i++){
        /* ch4 is not suspend/resume. bootup timer special */
        if(i == 4) continue;

        keep_oscr[i] = __REG(OSTIMER_REGS[i].oscr_addr);
        keep_osmr[i] = __REG(OSTIMER_REGS[i].osmr_addr);
        keep_omcr[i] = __REG(OSTIMER_REGS[i].omcr_addr);
    }

    return 0;
}

static int ostimer_resume(struct platform_device *dev)
{
    unsigned long irqsave;
    int i;

    pxa_set_cken(CKEN9_OSTIMER, 1); /* Enable OS timer clock */

    for(i = OSTIMER_CH_MIN; i <= OSTIMER_CH_MAX; i++){
        /* ch4 is not suspend/resume. bootup timer special */
        if(i == 4) continue;

        __REG(OSTIMER_REGS[i].omcr_addr) = keep_omcr[i];
        __REG(OSTIMER_REGS[i].osmr_addr) = keep_osmr[i];
        __REG(OSTIMER_REGS[i].oscr_addr) = keep_oscr[i];
    }

    spin_lock_irqsave(&ostimer_irq_lock, irqsave);
    OIER = (OIER & ~OSTIMER_VALID_BITMASK) | (keep_oier & OSTIMER_VALID_BITMASK);
    spin_unlock_irqrestore(&ostimer_irq_lock, irqsave);

    return 0;
}

#else
#define ostimer_suspend NULL
#define ostimer_resume NULL
#endif

static struct platform_driver ostimer_driver = {
    .suspend = ostimer_suspend,
    .resume = ostimer_resume,
    .driver = {
        .name = "xyz3-ostimer",
    }
};

/* +++++++++++++ File operations ++++++++++++++ */

static ssize_t ostimer_read(struct file* filp, char __user * buff,
                            size_t count, loff_t*  f_pos)
{
    ssize_t readsize = 0;
    char retbuff[15]; /* enough to store 32 bits */

    int ch = (int)(filp->private_data);
    unsigned long oscr = xyz3_ostimer_get_count(ch);
    int resolution_type = xyz3_ostimer_get_resolution(ch);

    /* head is not read */
    if(*f_pos != 0){ return 0; }

    if(resolution_type == OSTIMER_RESOLUTION_STOP){
        sprintf(retbuff, "disabled\n");
    }
    else {
        sprintf(retbuff, "%ld%s\n", oscr,
                (resolution_type == OSTIMER_RESOLUTION_1us)? "us" :
                ((resolution_type == OSTIMER_RESOLUTION_1ms)? "ms" : "s") );
    }

    readsize = strlen(retbuff);

    if(count < readsize){ return -EINVAL; }

    if(0 != copy_to_user(buff, retbuff, readsize) ){
        readsize = 0;
    }

    *f_pos += readsize;
    return readsize;
}

static int ostimer_ioctl(struct inode* inode, struct file* filp,
                         unsigned int cmd, unsigned long arg)
{
    unsigned long oscr;
    int ch = (int)(filp->private_data);
    xyz3_ostimer_interval_setting interval_setting;

    switch(cmd) {
    case OSTIMER_GET_COUNT:
        /* get count */
        oscr = xyz3_ostimer_get_count(ch);
        if(0 != copy_to_user((unsigned long*)arg, (unsigned long*)&oscr, sizeof(unsigned long)) ){
            return -EFAULT;
        }
        break;
    case OSTIMER_START_FREERUN:
        /* set count resolution and clear count */
        xyz3_ostimer_setting(ch, 0, 0, 0, arg);
        break;
    case OSTIMER_BLOCK_BY_COUNT:
        /* block by count match interrupt */
        xyz3_ostimer_blockcall(ch, arg, xyz3_ostimer_get_resolution(ch));
        break;
    case OSTIMER_SET_INTERVAL:
        /* set and start interval timer */
        if(0 != copy_from_user(&interval_setting, (void*)arg, sizeof(xyz3_ostimer_interval_setting)) ){
            return -EFAULT;
        }

        if(interval_setting.resolution != OSTIMER_RESOLUTION_STOP) {
            if(0 == xyz3_ostimer_register_callback(ch, interval_setting.count, interval_setting.resolution, ostimer_interval_handler_tbl[ch])){
                return -EBUSY;
            }
        }
        else {
            xyz3_ostimer_remove_callback(ch, ostimer_interval_handler_tbl[ch]);
        }

        break;
    case OSTIMER_LATCH_INTERVAL:
        /* wait occur interval timer */
        clear_bit(ch, &ostimer_interval_wakeup);
        wait_event_interruptible(ostimer_interval_wakeup_wq, test_bit(ch, &ostimer_interval_wakeup));

        break;
    default:
        return -ENOTTY;
    }

    return 0;
}

static int ostimer_open(struct inode* inode, struct file* filp)
{
    if((imajor(inode) != XYZ3_MAJOR_OSTIMER) ||
       ((iminor(inode) < XYZ3_MINOR_OSTIMER_MIN) || (XYZ3_MINOR_OSTIMER_MAX < iminor(inode))) ){
        return -ENODEV;
    }
  
    filp->private_data = (void*)(iminor(inode) - XYZ3_MINOR_OSTIMER_MIN + OSTIMER_CH_MIN);

    return 0;
}

static int ostimer_release(struct inode* inode, struct file* filp)
{
    /* Nothing process. */

    return 0;
}

static struct file_operations ostimer_fops = {
    owner: THIS_MODULE,
    read: ostimer_read,
    ioctl: ostimer_ioctl,
    open: ostimer_open,
    release: ostimer_release,
};
static struct cdev ostimer_cdev = {
    owner: THIS_MODULE,
    ops: &ostimer_fops,
};

/* +++++++++++++ End File operations ++++++++++++++ */

static int driver_registered = -1;
static int devno_registered = -1;
static int cdev_registered = -1;

#define XYZ3_MINOR_OSTIMER_NOOF (XYZ3_MINOR_OSTIMER_MAX - XYZ3_MINOR_OSTIMER_MIN +1)

static void xyz3_ostimer_exit( void )
{
    free_irq(IRQ_OST_4_11, NULL);

    if(cdev_registered == 0){
        cdev_del(&ostimer_cdev);
        cdev_registered = -1;
    }
    if(devno_registered == 0){
        unregister_chrdev_region(MKDEV(XYZ3_MAJOR_OSTIMER, XYZ3_MINOR_OSTIMER_MIN), XYZ3_MINOR_OSTIMER_NOOF);
        devno_registered = -1;
    }
    if(driver_registered == 0){
        platform_driver_unregister(&ostimer_driver);
        driver_registered = -1;
    }
}

static int __init xyz3_ostimer_init( void )
{
    pxa_set_cken(CKEN9_OSTIMER, 1); /* Enable OS timer clock */

    /* regist suspend resume support */
    driver_registered = platform_driver_register(&ostimer_driver);
    if(driver_registered != 0){
        xyz3_ostimer_exit();
        return -1;
    }

    /* regist DEVNO */
    devno_registered = register_chrdev_region(MKDEV(XYZ3_MAJOR_OSTIMER, XYZ3_MINOR_OSTIMER_MIN), XYZ3_MINOR_OSTIMER_NOOF, "ostimer");
    if(devno_registered != 0){
        xyz3_ostimer_exit();
        return -1;
    }
  
    /* regist CDEV */
    cdev_init(&ostimer_cdev, &ostimer_fops);
    ostimer_cdev.owner = THIS_MODULE;
    ostimer_cdev.ops = &ostimer_fops;
    cdev_registered = cdev_add(&ostimer_cdev, MKDEV(XYZ3_MAJOR_OSTIMER, XYZ3_MINOR_OSTIMER_MIN), XYZ3_MINOR_OSTIMER_NOOF);
    if(cdev_registered != 0){
        xyz3_ostimer_exit();
        return -1;
    }

    request_irq(IRQ_OST_4_11, ostimer_interrupt, SA_INTERRUPT, "OS timer 4 - 11 match interrupt", NULL);

    return 0;
}

module_init(xyz3_ostimer_init);
module_exit(xyz3_ostimer_exit);

MODULE_AUTHOR("");
MODULE_DESCRIPTION("Operating System Timers driver on PXA270");
MODULE_LICENSE("GPL");
