
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>

#include <asm/arch/system.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>

#include <asm/arch/xyz3ringbuf.h>

//#define DEBUG
#ifdef DEBUG
#include "xyz3ad7795.h"		// for debug
#define DPRN(x)		x
#else
#define DPRN(x)
#endif

#define BUF_UNREAD (-1)

Xyz3RingBuf *xyz3RingBufInit( int elements, int elementSize )
{
	Xyz3RingBuf *ringbuf;

	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	ringbuf = kmalloc( sizeof(Xyz3RingBuf), GFP_KERNEL );
	if( ringbuf == NULL )
	{
		printk( KERN_ERR "%s:kmalloc(Xyz3RingBuf) error!\n", __FUNCTION__ );
		return NULL;
	}
	memset( ringbuf, 0, sizeof(Xyz3RingBuf) );

	ringbuf->bufsize = elements * elementSize;

	ringbuf->buf = kmalloc( ringbuf->bufsize, GFP_KERNEL );
	if( ringbuf->buf == NULL )
	{
		printk( KERN_ERR "%s:kmalloc(ringbuf->buf) error!\n", __FUNCTION__ );
		kfree( ringbuf );
		return NULL;
	}

	ringbuf->elements = elements;
	ringbuf->elementSize = elementSize;
	ringbuf->wrPos = 0;
	ringbuf->rdPos = BUF_UNREAD;
	ringbuf->buflock = SPIN_LOCK_UNLOCKED;

	return ringbuf;
}

void xyz3RingBufFree( Xyz3RingBuf *ringbuf )
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );
	kfree( ringbuf->buf );
	kfree( ringbuf );
}


int xyz3RingBufWrite( Xyz3RingBuf *ringbuf, void *data )
{
#ifdef DEBUG
	PosSensor *ad = (PosSensor *)data;

	printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );
	printk( KERN_DEBUG "%s():wrPos=%d, rdPos=%d\n", __FUNCTION__, ringbuf->wrPos, ringbuf->rdPos );
	printk( KERN_DEBUG "%s():tick=0x%lx X=0x%x,Y=0x%x,Z=0x%x,T=0x%x,A=0x%x\n", __FUNCTION__, ad->tick, ad->g_x, ad->g_y, ad->g_z, ad->temp, ad->atmo );
#endif
	spin_lock( &ringbuf->buflock );

	memcpy( &ringbuf->buf[ringbuf->wrPos*ringbuf->elementSize],
	        data,
	        ringbuf->elementSize );

	ringbuf->wrPos ++;
	if( ringbuf->wrPos >= ringbuf->elements )
	{
		ringbuf->wrPos = 0;
		if( ringbuf->rdPos == BUF_UNREAD )
		{
			ringbuf->rdPos = 1;
		}
	}
	if( ringbuf->wrPos == ringbuf->rdPos )
	{
		ringbuf->rdPos ++;
		if( ringbuf->rdPos >= ringbuf->elements )
		{
			ringbuf->rdPos = 0;
		}
	}

	spin_unlock( &ringbuf->buflock );
	return 0;
}

/*
 * wrPos > rdPos -- none overlap.
 * wrPos < rdPos -- overlaped
 * wrPos = rdPos -- none data.
 */

int xyz3RingBufRead( Xyz3RingBuf *ringbuf, void *data )
{
	int readelements = 0;

	spin_lock( &ringbuf->buflock );

	if( ringbuf->rdPos == BUF_UNREAD )
	{
		if( ringbuf->wrPos > 0 )
		{
			ringbuf->rdPos = 0;
		}
	}

	// single element read
	if( ringbuf->rdPos != ringbuf->wrPos )
	{
		readelements = 1;

		memcpy( data, 
		        &ringbuf->buf[ringbuf->rdPos * ringbuf->elementSize],
		        ringbuf->elementSize );

		ringbuf->rdPos ++;
		if( ringbuf->rdPos >= ringbuf->elements )
		{
			ringbuf->rdPos = 0;
		}
	}

	spin_unlock( &ringbuf->buflock );
	return readelements;
}




EXPORT_SYMBOL(xyz3RingBufInit);
EXPORT_SYMBOL(xyz3RingBufFree);
EXPORT_SYMBOL(xyz3RingBufWrite);
EXPORT_SYMBOL(xyz3RingBufRead);
