/*
 * linux/arch/arm/mach-pxa/xyz4-serial.c
 *     -- Preset interval send on BTUart
 *        
 * Author:     Hirokazu ABE and Ryoichi KATO
 * Created:    May 23, 2007
 * Copyright:  Sony Corp.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
/* #include <linux/malloc.h> */
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/list.h>

#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/mach/irq.h>

#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>

#include <asm/arch/xyz4-serial.h>
#include <asm/arch/xyz3-devnum.h>

static spinlock_t sendbuff_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t recvbuff_lock = SPIN_LOCK_UNLOCKED;

/* send mode */
static atomic_t send_mode = ATOMIC_INIT(SERIAL_MODE_REALTIME);

/* interval send buff */
typedef struct {
    struct list_head list;
    uint8_t* data;
    int size;
} interval_sendbuff_t;

static LIST_HEAD(interval_sendbuff);
static LIST_HEAD(interval_sendbuff_trash);

/* ring buff */
#define sizeof_buffer 4096

typedef struct {
    uint8_t buff[sizeof_buffer];
    int head;
    int tail;
} ringbuff_t;

#define prev_pos(p) (((p) == 0)? (sizeof_buffer - 1): ((p)-1))
#define next_pos(p) (((p) == (sizeof_buffer - 1))? 0: ((p)+1))

static ringbuff_t sendbuff, recvbuff;

static int buff_empty(ringbuff_t* buff)
{
    return (buff->head == buff->tail)? 1 : 0;
}
static int put_buff(ringbuff_t* buff, uint8_t data)
{ 
    int ret = 0;

    if(prev_pos(buff->head) != buff->tail) {
        /* buffer is not full */
        (buff->buff)[buff->tail] = data;
        buff->tail = next_pos(buff->tail);
        ret = 1;
    }

    return ret;
}
static int get_buff(ringbuff_t* buff, uint8_t* data)
{
    int ret = 0;

    if(0 == buff_empty(buff)) {
        /* buffer is not empty */
        *data = (buff->buff)[buff->head];
        buff->head = next_pos(buff->head);
        ret = 1;
    }

    return ret;
}
static void clear_buff(ringbuff_t* buff)
{
    buff->head = buff->tail = 0;
}
static int buff_remain(ringbuff_t* buff)
{
    return (buff->tail + ((buff->head <= buff->tail) ? 0 : sizeof_buffer)) - buff->head;
}
static int buff_available(ringbuff_t* buff)
{
    return sizeof_buffer - buff_remain(buff) - 1;  /* head == tail is buffer empty */
}

/* --- BTUART Control --- */
#define SIZEOF_SENDFIFO 64
#define SIZEOF_RECVFIFO 64


static spinlock_t reg_lock = SPIN_LOCK_UNLOCKED;
static DECLARE_WAIT_QUEUE_HEAD(wait_recv_wq);

static void term_uart(void)
{
    unsigned long irqsave;
    volatile uint32_t syncread;

    spin_lock_irqsave(&reg_lock, irqsave);
    BTIER &= ~IER_UUE;
    syncread = BTIER; /* for split transaction */
    spin_unlock_irqrestore(&reg_lock, irqsave);
}
static void init_uart(int baudrate, int parity, int hwflow)
{
    unsigned long irqsave;
    unsigned long div, dll, dlh;

    if(baudrate == 0) return ;
    div = 14745600 / (16 * baudrate);
    dll = div & 0xFF;
    dlh = (div >> 8) & 0xFF;

    term_uart();

    spin_lock_irqsave(&reg_lock, irqsave);

    BTLCR = LCR_DLAB;
    BTDLH = dlh;
    BTDLL = dll;

    BTLCR = LCR_WLS1 | LCR_WLS0; /* 8bit */
    if(parity == SERIAL_PARITY_EVEN) {
        /* even parity */
        BTLCR |=  LCR_EPS | LCR_PEN;
    }
    else if(parity == SERIAL_PARITY_ODD) {
        /* odd parity */
        BTLCR |=  LCR_PEN;
    }
    /* else non parity */
    
    BTMCR = MCR_OUT2; /* enable interrupt */
    if(hwflow != 0) {
        /* enable hw flow control */
        BTMCR |= (1 << 5) /* MCR_AFE */ | MCR_RTS;
    }

    BTFCR = FCR_RESETTF | FCR_RESETRF | FCR_TRFIFOE;

    BTIER = IER_UUE | IER_RLSE | IER_RAVIE;

    spin_unlock_irqrestore(&reg_lock, irqsave);
}

static void start_tx(void)
{
    unsigned long irqsave;
    spin_lock_irqsave(&reg_lock, irqsave);
    BTIER |= IER_TIE;
    spin_unlock_irqrestore(&reg_lock, irqsave);
}
static void stop_tx(void)
{
    unsigned long irqsave;
    spin_lock_irqsave(&reg_lock, irqsave);
    BTIER &= ~IER_TIE;
    spin_unlock_irqrestore(&reg_lock, irqsave);
}

/* --- interval send --- */

static void serial_interval_handler(void)
{
    unsigned long irqsave;
    interval_sendbuff_t* buff_leaf;
    int i;
    
    if(list_empty(&interval_sendbuff)) return;

    spin_lock_irqsave(&sendbuff_lock, irqsave);

    buff_leaf = list_entry(interval_sendbuff.next, interval_sendbuff_t, list);
    if( buff_leaf->size < buff_available(&sendbuff) ) {
        for(i = 0; i < buff_leaf->size; i++) {
            put_buff(&sendbuff, (buff_leaf->data)[i]);
        }
        list_del(interval_sendbuff.next);

        list_add_tail(&(buff_leaf->list), &interval_sendbuff_trash); /* for remove in next write */

        spin_unlock_irqrestore(&sendbuff_lock, irqsave);

        start_tx();
    }
    else {
        spin_unlock_irqrestore(&sendbuff_lock, irqsave);
    }


}

static void clear_interval_sendbuff(struct list_head* buff)
{
    unsigned long irqsave;
    interval_sendbuff_t* buff_leaf;

    spin_lock_irqsave(&sendbuff_lock, irqsave);

    while(! list_empty(buff)) {

        buff_leaf = list_entry(buff->next, interval_sendbuff_t, list);
        list_del(buff->next);

        spin_unlock_irqrestore(&sendbuff_lock, irqsave);

        kfree(buff_leaf->data);
        kfree(buff_leaf);

        spin_lock_irqsave(&sendbuff_lock, irqsave);
    }

    INIT_LIST_HEAD(buff);

    spin_unlock_irqrestore(&sendbuff_lock, irqsave);
}

/* interrupt */

static irqreturn_t serial_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    irqreturn_t ret;
    unsigned long irqsave;

    int i = 0;
    uint32_t lsr;
    uint8_t sendchar;
    volatile uint32_t readerror;

    ret = IRQ_NONE;

    if(0 == (BTIIR & IIR_IP)) {
        lsr = BTLSR;

        /* Send chars */
        if(0 != (lsr & LSR_TDRQ)) {
            spin_lock_irqsave(&sendbuff_lock, irqsave);

            for(i = 0; i < SIZEOF_SENDFIFO/2; i++) {
                if(0 == get_buff(&sendbuff, &sendchar)) {
                    break;
                }
                BTTHR = sendchar;
            }

            if(0 != buff_empty(&sendbuff)) {
                stop_tx();
            }
    
            spin_unlock_irqrestore(&sendbuff_lock, irqsave);
        }

        /* Receive chars */
        while((lsr & LSR_DR) != 0) {
            if((BTIIR & (IIR_IID2 | IIR_IID1)) == 0x3) {
                /* RX error */
                if(lsr & LSR_FIFOE) printk(KERN_ERR "xyz4-serial : LSR_FIFOE\n");
                if(lsr & LSR_FE)    printk(KERN_ERR "xyz4-serial : LSR_FE\n");
                if(lsr & LSR_PE)    printk(KERN_ERR "xyz4-serial : LSR_PE\n");
                if(lsr & LSR_OE)    printk(KERN_ERR "xyz4-serial : LSR_OE\n");
                readerror = BTRBR;
            }
            else {
                spin_lock_irqsave(&recvbuff_lock, irqsave);
            
                if(0 == put_buff(&recvbuff, (uint8_t)BTRBR) ){
                    spin_unlock_irqrestore(&recvbuff_lock, irqsave);

                    printk(KERN_ERR "xyz4-serial : receive buffer overflow\n");

                    break;
                }

                spin_unlock_irqrestore(&recvbuff_lock, irqsave);

                wake_up_interruptible(&wait_recv_wq);
            }

            lsr = BTLSR;
        }


        ret = IRQ_HANDLED;
    }


    return ret;
}

#ifdef CONFIG_PM

static int serial_enabled = 0;
static int reserve_baudrate;
static int reserve_parity;
static int reserve_hwflow;

static int serial_suspend(struct platform_device *dev, pm_message_t state)
{
    if(serial_enabled) {
        term_uart();

        pxa_set_cken(CKEN7_BTUART, 0); /* Disable BT Uart clock */
    }

    return 0;
}

static int serial_resume(struct platform_device *dev)
{
    unsigned long irqsave;

    if(serial_enabled) {
        pxa_set_cken(CKEN7_BTUART, 1); /* Enable BT Uart clock */

        clear_interval_sendbuff(&interval_sendbuff);
        clear_interval_sendbuff(&interval_sendbuff_trash);

        spin_lock_irqsave(&sendbuff_lock, irqsave);
        clear_buff(&sendbuff);
        spin_unlock_irqrestore(&sendbuff_lock, irqsave);

        spin_lock_irqsave(&recvbuff_lock, irqsave);
        clear_buff(&recvbuff);
        spin_unlock_irqrestore(&recvbuff_lock, irqsave);

        init_uart(reserve_baudrate, reserve_parity, reserve_hwflow);
    }
    return 0;
}

#else
#define serial_suspend NULL
#define serial_resume NULL
#endif

static struct platform_driver serial_driver = {
    .suspend = serial_suspend,
    .resume = serial_resume,
    .driver = {
        .name = "xyz4-serial",
    }
};

/* +++++++++++++ File operations ++++++++++++++ */

static ssize_t serial_read(struct file* filp, char __user * buff,
                           size_t count, loff_t*  f_pos)
{
    unsigned long irqsave;

    ssize_t readsize;
    uint8_t* readdata;

    if(count == 0) return 0;

    wait_event_interruptible(wait_recv_wq, (0 < buff_remain(&recvbuff)) );

    readdata = kmalloc(count, GFP_KERNEL);
    if(readdata == NULL) {
        return -ENOMEM;
    }
    
    spin_lock_irqsave(&recvbuff_lock, irqsave);

    for(readsize = 0; readsize < count; readsize++) {
        if(0 == get_buff(&recvbuff, &(readdata[readsize])) ) {
            break;
        }
    }

    spin_unlock_irqrestore(&recvbuff_lock, irqsave);

    if(0 != copy_to_user(buff, readdata, readsize) ) {
        readsize = -EFAULT;
    }

    kfree(readdata);

    if(0 < readsize) {
        *f_pos += readsize;
    }
    return readsize;
}

static ssize_t serial_write(struct file* filp, const char __user * buff,
                            size_t count, loff_t*  f_pos)
{
    unsigned long irqsave;

    ssize_t writesize;
    int mode;
    interval_sendbuff_t* buff_leaf;

    uint8_t* writedata = kmalloc(count, GFP_KERNEL);
    if(writedata == NULL) {
        return -ENOMEM;
    }

    if(0 != copy_from_user(writedata, buff, count) ){
        kfree(writedata);
        return -EFAULT;
    }

    mode = atomic_read(&send_mode);

    if(mode == SERIAL_MODE_INTERVAL) {
        buff_leaf = kmalloc(sizeof(interval_sendbuff_t), GFP_KERNEL);
        if(buff_leaf == NULL) {
            kfree(writedata);
            return -ENOMEM;
        }

        buff_leaf->data = writedata;
        buff_leaf->size = count;
        writesize = count;
        
        spin_lock_irqsave(&sendbuff_lock, irqsave);

        list_add_tail(&(buff_leaf->list), &interval_sendbuff);

        spin_unlock_irqrestore(&sendbuff_lock, irqsave);

        clear_interval_sendbuff(&interval_sendbuff_trash); /* remove sended data */
    }
    else { /* mode == SERIAL_MODE_REALTIME */
        spin_lock_irqsave(&sendbuff_lock, irqsave);

        if(0 < buff_available(&sendbuff)) {
            for(writesize = 0; writesize < count; writesize++) {
                if(0 == put_buff(&sendbuff, writedata[writesize]) ) {
                    break;
                }
            }
        }
        else {
            writesize = -ENOSPC;
        }
        
        spin_unlock_irqrestore(&sendbuff_lock, irqsave);

        start_tx();

        kfree(writedata);
    }

    if(0 < writesize) {
        *f_pos += writesize;
    }
    return writesize;
}

static int serial_fsync(struct file* filp, struct dentry* dentp, int datasync)
{
    while(1) {
        if((BTLSR & LSR_TEMT) != 0) break;
        msleep(10);
    }

    return 0;
}

static int serial_ioctl(struct inode* inode, struct file* filp,
                         unsigned int cmd, unsigned long arg)
{
    unsigned long irqsave;
    xyz4_serial_port_setting port_setting;
    xyz4_serial_interval_setting interval_setting;

    switch(cmd) {
    case SERIAL_INIT:
        if(0 != copy_from_user((xyz4_serial_port_setting*)&port_setting,
                               (xyz4_serial_port_setting*)arg,
                               sizeof(xyz4_serial_port_setting)) ){
            return -EFAULT;
        }

#ifdef CONFIG_PM
        reserve_baudrate = port_setting.baud;
        reserve_parity = port_setting.parity;
        reserve_hwflow = port_setting.hwflow;
#endif

        clear_interval_sendbuff(&interval_sendbuff);
        clear_interval_sendbuff(&interval_sendbuff_trash);

        spin_lock_irqsave(&sendbuff_lock, irqsave);
        clear_buff(&sendbuff);
        spin_unlock_irqrestore(&sendbuff_lock, irqsave);

        spin_lock_irqsave(&recvbuff_lock, irqsave);
        clear_buff(&recvbuff);
        spin_unlock_irqrestore(&recvbuff_lock, irqsave);

        init_uart(port_setting.baud, port_setting.parity, port_setting.hwflow);

        break;
    case SERIAL_TERM:
        xyz3_ostimer_remove_callback(XYZ3_OSTIMER_CH_BTUART, &serial_interval_handler);

        term_uart();

        break;
    case SERIAL_SET_INTERVAL:
        if(0 != copy_from_user((xyz4_serial_interval_setting*)&interval_setting,
                               (xyz4_serial_interval_setting*)arg,
                               sizeof(xyz4_serial_interval_setting)) ){
            return -EFAULT;
        }

        if(interval_setting.resolution != OSTIMER_RESOLUTION_STOP){
            xyz3_ostimer_register_callback(XYZ3_OSTIMER_CH_BTUART,
                                           interval_setting.count, interval_setting.resolution,
                                           &serial_interval_handler);
        }
        else {
            xyz3_ostimer_remove_callback(XYZ3_OSTIMER_CH_BTUART, &serial_interval_handler);
        }

        break;
    case SERIAL_SET_MODE_INTERVAL:
        atomic_set(&send_mode, SERIAL_MODE_INTERVAL);

        break;
    case SERIAL_SET_MODE_REALTIME:
        atomic_set(&send_mode, SERIAL_MODE_REALTIME);

        break;

    default:
        return -ENOTTY;
    }

    return 0;
}

static int serial_open(struct inode* inode, struct file* filp)
{
    if((imajor(inode) != XYZ4_MAJOR_SERIAL) ||
       ((iminor(inode) != XYZ4_MINOR_SERIAL) && (iminor(inode) != XYZ4_MINOR_SERIAL_INTERVAL)) ){
        return -ENODEV;
    }

    if((BTIER & IER_UUE) != 0) {
        printk(KERN_ERR "xyz4-serial : Already enabled\n");
        return -EBUSY;
    }

#ifdef CONFIG_PM
    serial_enabled = 1;
#endif

    pxa_set_cken(CKEN7_BTUART, 1); /* Enable BT Uart clock */

    request_irq(IRQ_BTUART, serial_interrupt, SA_INTERRUPT, "Preset interval send on BTUART", NULL);

    INIT_LIST_HEAD(&interval_sendbuff);
    INIT_LIST_HEAD(&interval_sendbuff_trash);
    clear_buff(&sendbuff);
    clear_buff(&recvbuff);
    
    return 0;
}

static int serial_release(struct inode* inode, struct file* filp)
{
    xyz3_ostimer_remove_callback(XYZ3_OSTIMER_CH_BTUART, &serial_interval_handler);

    term_uart();

    free_irq(IRQ_BTUART, NULL);

#ifdef CONFIG_PM
    serial_enabled = 0;
#endif

    return 0;
}

static struct file_operations serial_fops = {
    owner: THIS_MODULE,
    read: serial_read,
    write: serial_write,
    fsync: serial_fsync,
    ioctl: serial_ioctl,
    open: serial_open,
    release: serial_release,
};
static struct cdev serial_cdev = {
    owner: THIS_MODULE,
    ops: &serial_fops,
};

/* +++++++++++++ End File operations ++++++++++++++ */

static int driver_registered = -1;
static int devno_registered = -1;
static int cdev_registered = -1;

static void xyz4_serial_exit( void )
{
    if(cdev_registered == 0){
        cdev_del(&serial_cdev);
        cdev_registered = -1;
    }
    if(devno_registered == 0){
        unregister_chrdev_region(MKDEV(XYZ4_MAJOR_SERIAL, XYZ4_MINOR_SERIAL), 2);
        devno_registered = -1;
    }
    if(driver_registered == 0){
        platform_driver_unregister(&serial_driver);
        driver_registered = -1;
    }
}

static int __init xyz4_serial_init( void )
{
    /* regist suspend resume support */
    driver_registered = platform_driver_register(&serial_driver);
    if(driver_registered != 0){
        xyz4_serial_exit();
        return -1;
    }

    /* regist DEVNO */
    devno_registered = register_chrdev_region(MKDEV(XYZ4_MAJOR_SERIAL, XYZ4_MINOR_SERIAL), 2, "Preset interval send on BTUART");
    if(devno_registered != 0){
        xyz4_serial_exit();
        return -1;
    }
  
    /* regist CDEV */
    cdev_init(&serial_cdev, &serial_fops);
    serial_cdev.owner = THIS_MODULE;
    serial_cdev.ops = &serial_fops;
    cdev_registered = cdev_add(&serial_cdev, MKDEV(XYZ4_MAJOR_SERIAL, XYZ4_MINOR_SERIAL), 2);
    if(cdev_registered != 0){
        xyz4_serial_exit();
        return -1;
    }

    return 0;
}

module_init(xyz4_serial_init);
module_exit(xyz4_serial_exit);

MODULE_AUTHOR("");
MODULE_DESCRIPTION("Preset interval send on BTUart");
MODULE_LICENSE("GPL");
