/*
 * xyz4-stealpcm.c - steal output pcm data from pcm_oss.c
 *
 * NOTE: Device's major / minor numbers are defined in <asm/arch/xyz3-devnum.h>
 *
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <asm/uaccess.h> /* copy_to_user(), copy_from_user() */

#include <asm/arch/xyz3-devnum.h>

/****************************************************************************/
/*
 * Prototypes
 */
static int stealpcm_open(struct inode * inode, struct file * filp);
static ssize_t stealpcm_read(struct file * file, char __user * buf,
                            size_t count, loff_t *ppos);
static int stealpcm_release(struct inode* inode, struct file* filp);

/****************************************************************************/
/*
 * Global data
 */

static struct file_operations stealpcm_fops = {
    .open       = stealpcm_open,
    .read       = stealpcm_read,
    .release    = stealpcm_release,
};

extern struct semaphore steal_pcm_data_lock;
extern uint32_t steal_pcm_data_size;
extern uint32_t steal_pcm_data_last;
extern uint8_t* steal_pcm_data;

/****************************************************************************/
/*
 * Functions
 */
static int stealpcm_open(struct inode * inode, struct file * filp)
{
    int ret = 0;

    if((imajor(inode) != XYZ4_MAJOR_STEALPCM) ||
       (iminor(inode) != XYZ4_MINOR_STEALPCM) ){
        return -ENODEV;
    }

    if(0 == down_interruptible(&steal_pcm_data_lock) ) {

        do {
            if(steal_pcm_data != NULL) {
                printk(KERN_INFO "stealpcm : already allocated\n");
                break;
            }

            steal_pcm_data_size = 48000 * 2 * 2 / 80; /* Max of data, 48kHz 16bit 2ch. 80Hz wave 1 cycle */
            steal_pcm_data = kmalloc(steal_pcm_data_size, GFP_KERNEL);
            steal_pcm_data_last = 0;
        
            if(steal_pcm_data == NULL) {
                steal_pcm_data_size = 0;
                ret = -ENOMEM;
                break;
            }

            memset(steal_pcm_data, 0, steal_pcm_data_size);

        } while(0);

        up(&steal_pcm_data_lock);
    }

    return ret;
}

static int stealpcm_release(struct inode* inode, struct file* filp)
{
    if(0 == down_interruptible(&steal_pcm_data_lock) ) {

        kfree(steal_pcm_data);
        steal_pcm_data = NULL;
        steal_pcm_data_size = 0;
        steal_pcm_data_last = 0;

        up(&steal_pcm_data_lock);
    }
    return 0;
}

static ssize_t stealpcm_read(struct file * file, char __user * buf,
                            size_t count, loff_t *ppos)
{
    ssize_t ret = 0;

    if ( *ppos != 0 ) {
        /* already written; return EOF */
        return 0;
    }
    else if(steal_pcm_data != NULL) {
        int32_t copy_size, total_size, copy_start;
        total_size = (steal_pcm_data_size <= (int32_t)count) ? steal_pcm_data_size : count;
    
        copy_start = steal_pcm_data_last - total_size;
        if(copy_start < 0) {
            copy_start += steal_pcm_data_size;
            copy_size = steal_pcm_data_size - copy_start;
            copy_to_user(buf, &(steal_pcm_data[copy_start]), copy_size);
            copy_to_user(&(buf[copy_size]), steal_pcm_data, total_size - copy_size);
        }
        else {
            copy_to_user(buf, &(steal_pcm_data[copy_start]), total_size);
        }

        ret = total_size;
    }
    return ret;
}

/****************************************************************************/

/*
 * Initialize / Finalize Routines
 */
static void __exit stealpcm_exit( void )
{
    unregister_chrdev(XYZ4_MAJOR_STEALPCM,"stealpcm");
}

static int __init stealpcm_init( void )
{
    if (register_chrdev(XYZ4_MAJOR_STEALPCM,"stealpcm",&stealpcm_fops)) {
        printk(KERN_ERR "stealpcm: Unable to get major %d for memory devs.\n",
               XYZ4_MAJOR_STEALPCM);
        return -ENODEV;
    }
    return 0;
}

module_init(stealpcm_init);
module_exit(stealpcm_exit);

