
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/interrupt.h>

#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/mach/irq.h>
#include <asm/arch/irqs.h>

#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>

#define noof_array(commands) (sizeof(commands) / sizeof(commands[0]))

typedef struct {
    uint32_t reg_no;
    uint32_t data;
} lcdssp_command_t;

typedef struct {
    const lcdssp_command_t* commands;
    int noof_commands;
    int post_wait;
} lcdssp_command_seq_t;

/* Power and Display ON command set */
static const lcdssp_command_t lcdssp_commands_on_1[] = {
    /* Power On Sequence */
    { 0x11, 0x0010},
    { 0x12, 0x001a},
    { 0x13, 0x0610},
    { 0x14, 0x0000},
    { 0x15, 0x03cd},
    { 0x16, 0x0000},
    { 0x18, 0x7705},
    { 0x01, 0x4000},
    { 0x02, 0x0300},
    { 0x03, 0x0000},
    { 0x04, 0x0001},
    { 0x05, 0x0006}, /* VBP = 6 */
    { 0x06, 0x0008}, /* HBP = 8 */
    { 0x08, 0x0004},
    { 0x0b, 0x0200},
    { 0x0c, 0x0200},
    { 0x0e, 0x0000},
    { 0x0f, 0x0000},
    { 0x20, 0x0000},
    { 0x21, 0x0700},
    { 0x2a, 0x0000},
    { 0x2b, 0x0000},
    { 0x10, 0x0004},
    { 0x30, 0x0b07},
    { 0x31, 0x0b07},
    { 0x32, 0x0707},
    { 0x33, 0x0707},
    { 0x34, 0x0707},
    { 0x35, 0x0707},
    { 0x36, 0x0701},
    { 0x37, 0x0701},
    { 0x38, 0x1a1f},
    { 0x39, 0x000a},
};
static const lcdssp_command_t lcdssp_commands_on_2[] = {
    /* Power On Sequence. continuation */
    { 0x11, 0x0710},
    /* Display On Sequence */
    { 0x07, 0x0001},
};
static const lcdssp_command_t lcdssp_commands_on_3[] = {
    /* Display On Sequence. continuation */
    { 0x07, 0x0101},
};
static const lcdssp_command_t lcdssp_commands_on_4[] = {
    /* Display On Sequence. continuation */
    { 0x07, 0x010b},
};
static const lcdssp_command_seq_t lcdssp_commands_on[] = {
    {lcdssp_commands_on_1, noof_array(lcdssp_commands_on_1), 100},
    {lcdssp_commands_on_2, noof_array(lcdssp_commands_on_2), 10},
    {lcdssp_commands_on_3, noof_array(lcdssp_commands_on_3), 10},
    {lcdssp_commands_on_4, noof_array(lcdssp_commands_on_4), 0},
};

/* Power and Display OFF command set */
static lcdssp_command_t lcdssp_commands_off_1[] = {
    /* Display Off Sequence */
    {0x0B, 0x3020},
    {0x07, 0x0102},
};
static lcdssp_command_t lcdssp_commands_off_2[] = {
    {0x07, 0x0000},
};
static lcdssp_command_t lcdssp_commands_off_3[] = {
    /* Power Off Sequence */
    {0x10, 0x0000},
    {0x11, 0x0010},
};
static const lcdssp_command_seq_t lcdssp_commands_off[] = {
    {lcdssp_commands_off_1, noof_array(lcdssp_commands_off_1), 10},
    {lcdssp_commands_off_2, noof_array(lcdssp_commands_off_2), 10},
    {lcdssp_commands_off_3, noof_array(lcdssp_commands_off_3), 0},
};


static DECLARE_WAIT_QUEUE_HEAD(lcdssp_transmitted_wq);

#define SSSR_TFL(SSSR_Px) (((SSSR_Px) >> 8) & 0xf)
#define SSSR_T_FIFO_EMPTY(SSSR_Px) ((SSSR_TFL(SSSR_Px) == 0) && (((SSSR_Px) & SSSR_TNF) != 0))

static int lcdssp_trans_commands(const lcdssp_command_t* commands, int noof)
{
    int ret = 0;
    int index = 0;
    int i;
    int timeout;
    uint32_t transdata[2];

    while (index < noof) {
        transdata[0] = 0x640000 | commands[index].reg_no;
        transdata[1] = 0x660000 | commands[index].data;

        for(i = 0; i < 2; i++) {
            SSDR_P3 = transdata[i];

            SSCR1_P3 |= (SSCR1_TxTresh(8) & SSCR1_TFT) + SSCR1_TIE;

            timeout = wait_event_interruptible_timeout(lcdssp_transmitted_wq, (SSSR_P3 & SSSR_TNF) != 0, HZ);
            if(timeout == 0) {
                /* timeout */
                printk(KERN_ERR "sonynavu-lcdssp : error transmit command timeout.\n");
                ret = -EIO;
                goto func_out;
            }
            else if(timeout < 0){
                /* signal interrupt */
                ret = -EINTR;
                goto func_out;
            }
        }

        index++;
    }

    /* wait transmitted all data */

    SSCR0_P3 &= ~SSCR0_TIM; /* T FIFO underrun interrupt unmask */

    timeout = wait_event_interruptible_timeout(lcdssp_transmitted_wq, SSSR_T_FIFO_EMPTY(SSSR_P3), HZ);
    if(timeout == 0 ) {
        /* timeout */
        printk(KERN_ERR "sonynavu-lcdssp : error not respond wait transmitted all data.\n");
        ret = -EIO;
        goto func_out;
    }
    else if(timeout < 0) {
        /* signal interrupt */
        ret = -EINTR;
        goto func_out;
    }

  func_out:

    return ret;
}

static irqreturn_t lcdssp_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    if((SSSR_P3 & (SSSR_TFS | SSSR_TUR)) == 0) {
        return IRQ_NONE;
    }

    if((SSSR_P3 & SSSR_TFS) != 0) {
        SSCR1_P3 &= (~SSCR1_TFT & ~SSCR1_TIE);
    }
    if((SSSR_P3 & SSSR_TUR) != 0) {
        SSSR_P3 = SSSR_TUR;
        SSCR0_P3 |= SSCR0_TIM;
    }
    wake_up_interruptible(&lcdssp_transmitted_wq);

    return IRQ_HANDLED;
}

static int lcdssp_trans_commands_in_atomic(const lcdssp_command_t* commands, int noof)
{
    int ret = 0;
    int index = 0;
    int i;
    int timeout;
    uint32_t transdata[2];

    while (index < noof) {
        transdata[0] = 0x640000 | commands[index].reg_no;
        transdata[1] = 0x660000 | commands[index].data;

        for(i = 0; i < 2; i++) {
            SSDR_P3 = transdata[i];

            for(timeout = 1000; timeout <= 0; timeout--) {
                if( (SSSR_P3 & SSSR_TNF) != 0 ) {
                    break;
                }
                mdelay(1);
            }
            if(timeout <= 0) {
                /* timeout */
                printk(KERN_ERR "sonynavu-lcdssp : error transmit command timeout.\n");
                ret = -EIO;
                goto func_out;
            }
        }

        index++;
    }

    /* wait transmitted all data */

    for(timeout = 1000; timeout <= 0; timeout--) {
        if( SSSR_T_FIFO_EMPTY(SSSR_P3) ) {
            break;
        }
        mdelay(1);
    }
    if(timeout <= 0) {
        /* timeout */
        printk(KERN_ERR "sonynavu-lcdssp : error not respond wait transmitted all data.\n");
        ret = -EIO;
        goto func_out;
    }

  func_out:

    return ret;
}

#define SSCR0_SCR_DIV(x) (((x)-1) << 8) /* SCR ( 13Mbps / x [1..4096] ) */
#define SSCR0_E_DSS(x) (((((x)-1) & 0x10) << (20-4)) + (((x)-1) & SSCR0_DSS))

static int lcdssp_control_on_off(int on)
{
    int ret;
    int i;

    pxa_set_cken(CKEN4_SSP3, 1); /* Enable SSP3 clock */

    SSCR0_P3 = 
        SSCR0_TIM + SSCR0_RIM + /* T FIFO underrun / R FIFO overrun interrupt mask */
        (SSCR0_SCR_DIV(1) & SSCR0_SCR) + /* SCR  13Mbps */
        (SSCR0_Motorola & SSCR0_FRF) + /* FRF  Motorola SPI format */
        (SSCR0_E_DSS(24) & (SSCR0_EDSS | SSCR0_DSS)); /* EDSS, DSS  24bit data */

    SSCR1_P3 = 0; /* Disable transmit FIFO interrupt etc. initial value */
    
    if(! in_atomic() ) {
        request_irq(IRQ_SSP3, lcdssp_interrupt, 0, "SSP for LCD Panel interrupt", NULL);
    }

    SSCR0_P3 |= SSCR0_SSE; /* Enable SSP */

    /* ON */
    if(on != 0) {
        /* enter power and display on sequence */
        for(i = 0; i < noof_array(lcdssp_commands_on); i++) {
            if( in_atomic() ) {
                ret = lcdssp_trans_commands_in_atomic(lcdssp_commands_on[i].commands, lcdssp_commands_on[i].noof_commands);
                if(ret != 0) {
                    break;
                }
                mdelay(lcdssp_commands_on[i].post_wait);
            }
            else {
                ret = lcdssp_trans_commands(lcdssp_commands_on[i].commands, lcdssp_commands_on[i].noof_commands);
                if(ret != 0) {
                    break;
                }
                if(0 != msleep_interruptible(lcdssp_commands_on[i].post_wait)) {
                    ret = -EINTR;
                    break;
                }
            }
        }
    }
    /* OFF */
    else {
        /* enter power and display off sequence */
        for(i = 0; i < noof_array(lcdssp_commands_off); i++) {
            if( in_atomic() ) {
                ret = lcdssp_trans_commands_in_atomic(lcdssp_commands_off[i].commands, lcdssp_commands_off[i].noof_commands);
                if(ret != 0) {
                    break;
                }
                mdelay(lcdssp_commands_off[i].post_wait);
            }
            else {
                ret = lcdssp_trans_commands(lcdssp_commands_off[i].commands, lcdssp_commands_off[i].noof_commands);
                if(ret != 0) {
                    break;
                }
                if(0 != msleep_interruptible(lcdssp_commands_off[i].post_wait)) {
                    ret = -EINTR;
                    break;
                }
            }
        }
    }

    SSCR0_P3 &= ~SSCR0_SSE; /* Disable SSP */

    if(! in_atomic() ) {
        free_irq(IRQ_SSP3, NULL);
    }

    SSCR0_P3 = SSCR0_TIM + SSCR0_RIM; /* T FIFO underrun / R FIFO overrun interrupt mask */
    SSCR1_P3 = 0; /* Disable transmit FIFO interrupt etc. initial value */

    return ret;
}

int sonynavu_lcdssp_power_on(void)
{
    return lcdssp_control_on_off(1);
}
int sonynavu_lcdssp_power_off(void)
{
    return lcdssp_control_on_off(0);
}
EXPORT_SYMBOL(sonynavu_lcdssp_power_on);
EXPORT_SYMBOL(sonynavu_lcdssp_power_off);
