#ifndef __NAVU_GPIO_H
#define __NAVU_GPIO_H

/****************************************************************************
 *             IOCTL Interface Defines for User Space Programs
 ****************************************************************************/
#ifdef __cplusplus
extern "C" {
#endif/*__cplusplus*/
#include <linux/ioctl.h>

#define NAVU_IRQ_FALL_EDGE          0x0001
#define NAVU_IRQ_RISE_EDGE          0x0002
#define NAVU_IRQ_DOUBLE_EDGE        (NAVU_IRQ_FALL_EDGE|NAVU_IRQ_RISE_EDGE)
#define NAVU_IRQ_LOW_LEVEL          0x0010
#define NAVU_IRQ_HIGH_LEVEL         0x0020


typedef struct {
    unsigned long   port;    /* GPIO Port name */
    unsigned int    trigger; /* Wait trigger type */
    char            *name;   /* Deprecated; not used anymore */
} navu_irq_ioctl_t;

/*
 * GET_LEVEL: Get current state of a GPIO port
 * SET_HIGH:  Set a GPIO port to high
 * SET_LOW    Set a GPIO port to low
 * IRQ_WAIT   Wait for a GPIO interrupt, port and trigger specified by
 *            navu_irq_ioctl_t arg. The ioctl(2) call should block and
 *            the calling user-space program should sleep until
 *            requested interrupt is delivered.
 * IRQ_CANCEL Force cancel waiting by IRQ_WAIT ioctl() call.
 *            ALL processes waiting for a specified GPIO port should be
 *            waken up, regardress of permission or user-ID.
 *            (so, this can be a security issue if improperly used)
 */
#define NAVUGP_MAGIC 'k'
#define NAVUGP_GET_LEVEL    (_IOR(NAVUGP_MAGIC, 1, sizeof(unsigned long)))
#define NAVUGP_SET_HIGH     (_IOW(NAVUGP_MAGIC, 2, sizeof(unsigned long)))
#define NAVUGP_SET_LOW      (_IOW(NAVUGP_MAGIC, 3, sizeof(unsigned long)))
#define NAVUGP_IRQ_WAIT     (_IOW(NAVUGP_MAGIC, 4, sizeof(navu_irq_ioctl_t)))
#define NAVUGP_IRQ_CANCEL   (_IOW(NAVUGP_MAGIC, 5, sizeof(unsigned long)))



#define NAVU_GPIO_KERNEL_H_IS_KERNEL_ONLY
#include "navu-gpio-kernel.h"
#undef  NAVU_GPIO_KERNEL_H_IS_KERNEL_ONLY

#ifndef __KERNEL__
    /*
     * Though it's not smart to expose kernel config to user programs,
     * I don't have good alternative...
     */
#    include <linux/autoconf.h>
#endif/*__KERNEL__*/


/****************************************************************************
 *                    GPIO Symbolic Names for XYZ3
 ****************************************************************************/
#ifdef CONFIG_MACH_SONY_NAVU_XYZ3
#    define MACH_XYZ3
#    define NAVU_IN_POWER_KEY           NAVU_GPIO_0     /* H:On */
#    define NAVU_IN_ACC_IRQ             NAVU_GPIO_1     /* L:On*/
#    define NAVU_IN_AC97_IRQ            NAVU_GPIO_13    /* AC97 Interrupt */
#    define NAVU_IN_GPS_PPS             NAVU_GPIO_17    /* L -> H */
#    define NAVU_IN_BATT_CHG            NAVU_GPIO_20    /* L: Charging */
#    define NAVU_IN_CF_IRQ              NAVU_GPIO_22
#    define NAVU_IN_USB_DETECT          NAVU_GPIO_27    /* L: Connected */
#    define NAVU_IN_ON_CRADLE           NAVU_GPIO_41    /* L: Connected */
#    define NAVU_IN_PARKING             NAVU_GPIO_53    /* L: Parking */
#    define NAVU_IN_DOC_IRQ             NAVU_GPIO_110   /* L: Active */
#    define NAVU_IN_MS_INS              NAVU_GPIO_112   /* L: Insert */
#    define NAVU_IN_RESET_DET           NAVU_GPIO_115   /* H: Reset detect */
#    define NAVU_OUT_LCD25V_ON          NAVU_GPIO_11    /* H: LCD2.5V ON */
#    define NAVU_OUT_LCDON              NAVU_GPIO_14    /* H: ON */
#    define NAVU_OUT_NAVI_ATT           NAVU_GPIO_19    /* H: Att */
#    define NAVU_OUT_LCD5V_ON           NAVU_GPIO_35    /* H: LCD 5V On */
#    define NAVU_OUT_FLASH_WP           NAVU_GPIO_82    /* L: Protect */
#    define NAVU_OUT_GPS_RESET          NAVU_GPIO_90    /* L: Reset */
#    define NAVU_OUT_GPS_RAM_CLR        NAVU_GPIO_91    /* H: Clear */
#    define NAVU_OUT_USB_DPLUS_ON       NAVU_GPIO_96    /* H: ON */
#    define NAVU_OUT_CF_POWER_EN        NAVU_GPIO_97    /* H: Enable */
#    define NAVU_OUT_CF_RESET           NAVU_GPIO_99    /* H: Reset */
#    define NAVU_OUT_CF_ADDR_EN         NAVU_GPIO_100   /* H: Enable */
#    define NAVU_OUT_MS_POWER_EN        NAVU_GPIO_101   /* H: Enable */
#    define NAVU_OUT_BT_POWER_EN        NAVU_GPIO_104   /* H: Enable */
#    define NAVU_OUT_RDS_POWER_EN       NAVU_GPIO_106   /* H: Enable */
#    define NAVU_OUT_SP_MUTE            NAVU_GPIO_111   /* L: Muting */
#endif


/****************************************************************************
 *                    GPIO Symbolic Names for XYZ4
 ****************************************************************************/
#ifdef CONFIG_MACH_SONY_NAVU_XYZ4
#    define MACH_XYZ4
/*
 * L: Low, H: High,  @: New GPIO (from xyz3),  %: GPIO Port Changed from xyz3
 */
/* GPIO Input */
#    define NAVU_IN_POWER_KEY      NAVU_GPIO_0   /* H:Power key pressed */
#    define NAVU_IN_ACC_IRQ        NAVU_GPIO_1   /* L:ACC power on */
//#  define NAVU_IN_LAN_IRQ        NAVU_GPIO_10  /* Ethernet IRQ */
#    define NAVU_IN_VOICE_SW       NAVU_GPIO_22  /*@H: Voice button on */
#    define NAVU_IN_HP_DET         NAVU_GPIO_12  /*@L:Headphone on */
#    define NAVU_IN_AC97_IRQ       NAVU_GPIO_13
#    define NAVU_IN_GPS_PPS        NAVU_GPIO_17  /* GPS 1PPS input */
#    define NAVU_IN_BATT_CHG       NAVU_GPIO_20  /* L: Battery is charging */
#    define NAVU_IN_USB_DETECT     NAVU_GPIO_27  /* L: Connected */
#    define NAVU_IN_ON_CRADLE      NAVU_GPIO_41  /* L: Connected */
#    define NAVU_IN_RDS_xIRQ       NAVU_GPIO_53  /*@L:Interrupted*/
#    define NAVU_IN_MS_INS         NAVU_GPIO_57  /*%L:MemoryStick inserted*/
#    define NAVU_IN_MOVINAND_JIG   NAVU_GPIO_107 /*@L:MoviNAND writer connected*/
#    define NAVU_IN_USB_JIG        NAVU_GPIO_115 /*@L:USB mini-A jig inserted */
/* GPIO Output */
#    define NAVU_OUT_LCD25V_ON     NAVU_GPIO_11  /* H:LCD2.5V ON */
#    define NAVU_OUT_LCDON         NAVU_GPIO_14  /* H:ON */
#    define NAVU_OUT_MS_LED_EN     NAVU_GPIO_19  /*@L:MemoryStick LED Force OFF*/
#    define NAVU_OUT_FLASH_WP      NAVU_GPIO_21  /* L:NOR flash Write protect */
#    define NAVU_OUT_MSC_POWER     NAVU_GPIO_35  /*@H: Enable MemoryStick Controller */
//#  define NAVU_IN_WLAN_INT       NAVU_GPIO_36  /*%WLAN IRQ */
#    define NAVU_OUT_WLAN_POWER    NAVU_GPIO_37  /* L:WLAN Power off */
#    define NAVU_OUT_WLAN_RESET    NAVU_GPIO_38  /* L:WLAN reset */
#    define NAVU_OUT_SP_MUTE       NAVU_GPIO_40  /*%L:SP1 Muting */
#    define NAVU_OUT_BATT_KILL     NAVU_GPIO_88  /*@H:Kill. Same as reset button */
#    define NAVU_OUT_GPS_RESET     NAVU_GPIO_90  /* L:Reset */
#    define NAVU_OUT_GPS_RAM_CLR   NAVU_GPIO_91  /* H:Clear */
#    define NAVU_OUT_USB_DPLUS_ON  NAVU_GPIO_96  /* H:Pull up D+*/
#    define NAVU_OUT_HP_MUTE       NAVU_GPIO_97  /*@L:Headphone mute */
#    define NAVU_OUT_ANT_SEL_P     NAVU_GPIO_99  /* Select antenna; L=WLAN, H=BT */
#    define NAVU_OUT_ANT_SEL_N     NAVU_GPIO_100 /* Select antenna; L=BT, H=WLAN */

#    define NAVU_OUT_TP_LED_ON     NAVU_GPIO_101 /*@H:Touch-panel backlight LED on*/
#    define NAVU_OUT_BT_POWER      NAVU_GPIO_104 /*@L:BT module power AND LED off*/
#    define NAVU_OUT_RDS_POWER     NAVU_GPIO_106 /*@L:RDS module power off */

#endif




#ifdef __cplusplus
}
#endif/*__cplusplus*/
#endif/*__NAVU_GPIO_H */

